---
---  David J. Sankel (camio@yahoo.com) has simply commented out some lines
---  in this file to make it a generalizable toolkit.  Taken from the 
---  htoolkit sourceforge repository.
---
-----------------------------------------------------------------------------------------
{-| Module      :  Attributes
    Copyright   :  (c) Daan Leijen 2003
    License     :  BSD-style

    Maintainer  :  daan@cs.uu.nl
    Stability   :  provisional
    Portability :  portable

   Widgets @w@ can have attributes of type @a@ represented by the type @Attr w a@.
   An attribute set or read. An example of an attribute is @title@ with type:

   > title :: Attr Window String

   When an attribute is associated with a value, we call it a /property/ of type @Prop w@.
   Properties are constructed with the ('=:') function:

   > title =: "hello world"  :: Prop Window

   Properties can be set with the 'set' function:

   > set win [title =: "hello world"]  :: IO ()

   and attributes can be read with the 'get' function:

   > get win title  :: IO String

   The function 'get', 'set' and '(=:)' are polymorphic and work for all widgets, but
   the @title@ attribute just works for windows. Many attributes are defined for multiple
   widgets and are organised in type classes, for example 'Widget' and 'Dimensions'.

   The ('~:') operator is used to transform an attribute with an update function.
   For example, the 'interval' on a timer can be doubled with:

   > set timer [interval ~: (*2)]
-}
-----------------------------------------------------------------------------------------
module Attributes
             ( 
             -- * Attributes and properties
               Attr, Prop
             , set, get, with, (=:), (~:), (=::), (~::)
             , mapAttr
             , newAttr
             ) where

--import Graphics.UI.GIO.Types

infixr 0 =:, =::
infixr 0 ~:, ~::

{--------------------------------------------------------------------
  Attributes
--------------------------------------------------------------------}
-- | Widgets @w@ can have attributes of type @a@. 
data Attr w a = Attr (w -> IO a) (w -> a -> IO ())  

-- | Create a new attribute with a specified getter and setter function.
newAttr :: (w -> IO a) -> (w -> a -> IO ()) -> Attr w a
newAttr getter setter
  = Attr getter setter

-- | A property of a widget @w@ is an attribute that
-- is already associated with a value. Properties are
-- constructed with the '(=:)' operator.
data Prop w   = Prop{ propSet     :: (w -> IO ())
                    , propRestore :: (w -> IO (IO ()))  -- a function that returns a restoration function :-)
                    }


-- | (@mapAttr get set attr@) maps an attribute of @Attr w a@ to
-- @Attr w b@ where (@get :: a -> b@) is used when the attribute is
-- requested and (@set :: a -> b -> a@) is applied to current
-- value when the attribute is set.
mapAttr :: (a -> b) -> (a -> b -> a) -> Attr w a -> Attr w b
mapAttr get set (Attr getter setter)
    = Attr (\w   -> do a <- getter w; return (get a))
           (\w b -> do a <- getter w; setter w (set a b))

-- | Set properties
set :: w -> [Prop w] -> IO ()
set w props
  = mapM_ (set1 w) props

set1 :: w -> Prop w -> IO ()
set1 w (Prop setter mkrestore) 
  = setter w

-- | Get an attribute
get :: w -> Attr w a -> IO a
get w (Attr getter setter)
  = getter w

-- | Set properties just for a certain computation. The previous values are automatically
-- restored on return.
with :: w -> [Prop w] -> IO a -> IO a
with w props io
  = foldr (with1 w) io props

with1 :: w -> Prop w -> IO a -> IO a
with1 w (Prop setter mkrestore) io
  = do restore <- mkrestore w
       setter w 
       x <- io
       restore
       return x

-- | Associate an attribute with a new value into a property.
(=:) :: Attr w a -> a -> Prop w
(=:) (Attr getter setter) x
  = Prop (\w -> setter w x) 
         (\w -> do oldx <- getter w; return (setter w oldx))
          
-- | Apply an update function to an attribute.
(~:) :: Attr w a -> (a -> a) -> Prop w
(~:) (Attr getter setter) f
  = Prop (\w -> do x <- getter w; setter w (f x))
         (\w -> do oldx <- getter w; return (setter w oldx))

-- | Set the value of an attribute with a function that takes the widget
-- itself as an argument.
(=::) :: Attr w a -> (w -> a) -> Prop w
(=::) (Attr getter setter) f
  = Prop (\w -> setter w (f w))
         (\w -> do oldx <- getter w; return (setter w oldx))
          
-- | Set the value of an attribute with a function that takes the widget
-- itself and the current value of the attribute as arguments. 
(~::) :: Attr w a -> (w -> a -> a) -> Prop w
(~::) (Attr getter setter) f
  = Prop (\w -> do x <- getter w; setter w (f w x))
         (\w -> do oldx <- getter w; return (setter w oldx))
